@@ -45,6 +45,8 @@
 
 #ifdef HAVE_OPENSSL
 #include <openssl/rand.h>
+#elif defined(HAVE_GNUTLS)
+#include <gcrypt.h>
 #endif
 
 #include <errno.h>
@@ -107,9 +109,7 @@ struct auth_handler {
 struct auth_challenge {
     const struct auth_protocol *protocol;
     struct auth_handler *handler;
-    const char *realm;
-    const char *nonce;
-    const char *opaque;
+    const char *realm, *nonce, *opaque, *domain;
     unsigned int stale; /* if stale=true */
     unsigned int got_qop; /* we were given a qop directive */
     unsigned int qop_auth; /* "auth" token in qop attrib */
@@ -178,6 +178,8 @@ typedef struct {
     char *nonce;
     char *cnonce;
     char *opaque;
+    char **domains; /* list of paths given as domain. */
+    size_t ndomains; /* size of domains array */
     auth_qop qop;
     auth_algorithm alg;
     unsigned int nonce_count;
@@ -244,6 +246,16 @@ struct auth_protocol {
 static void challenge_error(ne_buffer **errmsg, const char *fmt, ...)
     ne_attribute((format(printf, 2, 3)));
 
+/* Free the domains array, precondition sess->ndomains > 0. */
+static void free_domains(auth_session *sess)
+{
+    do {
+        ne_free(sess->domains[sess->ndomains - 1]);
+    } while (--sess->ndomains);
+    ne_free(sess->domains);
+    sess->domains = NULL;
+}
+
 static void clean_session(auth_session *sess) 
 {
     if (sess->basic) ne_free(sess->basic);
@@ -257,6 +269,7 @@ static void clean_session(auth_session *sess)
         ne_md5_destroy_ctx(sess->stored_rdig);
         sess->stored_rdig = NULL;
     }
+    if (sess->ndomains) free_domains(sess);
 #ifdef HAVE_GSSAPI
     {
         unsigned int major;
@@ -285,43 +298,46 @@ static char *get_cnonce(void)
 
     hash = ne_md5_create_ctx();
 
-#ifdef HAVE_OPENSSL
-    if (RAND_status() == 1 && RAND_pseudo_bytes(data, sizeof data) >= 0)
+#ifdef HAVE_GNUTLS
+    if (1) {
+        gcry_create_nonce(data, sizeof data);
+        ne_md5_process_bytes(data, sizeof data, hash);
+    }
+    else
+#elif defined(HAVE_OPENSSL)
+    if (RAND_status() == 1 && RAND_pseudo_bytes(data, sizeof data) >= 0) {
 	ne_md5_process_bytes(data, sizeof data, hash);
-    else {
-#endif
-    /* Fallback sources of random data: all bad, but no good sources
-     * are available. */
-
-    /* Uninitialized stack data; yes, happy valgrinders, this is
-     * supposed to be here. */
-    ne_md5_process_bytes(data, sizeof data, hash);
-    
+    } 
+    else 
+#endif /* HAVE_OPENSSL */
+    {
+        /* Fallback sources of random data: all bad, but no good sources
+         * are available. */
+        
+        /* Uninitialized stack data; yes, happy valgrinders, this is
+         * supposed to be here. */
+        ne_md5_process_bytes(data, sizeof data, hash);
+        
+        {
 #ifdef HAVE_GETTIMEOFDAY
-    {
-	struct timeval tv;
-	if (gettimeofday(&tv, NULL) == 0)
-	    ne_md5_process_bytes(&tv, sizeof tv, hash);
-    }
+            struct timeval tv;
+            if (gettimeofday(&tv, NULL) == 0)
+                ne_md5_process_bytes(&tv, sizeof tv, hash);
 #else /* HAVE_GETTIMEOFDAY */
-    {
-	time_t t = time(NULL);
-	ne_md5_process_bytes(&t, sizeof t, hash);
-    }
+            time_t t = time(NULL);
+            ne_md5_process_bytes(&t, sizeof t, hash);
 #endif
-    {
+        }
+        {
 #ifdef WIN32
-	DWORD pid = GetCurrentThreadId();
+            DWORD pid = GetCurrentThreadId();
 #else
-	pid_t pid = getpid();
+            pid_t pid = getpid();
 #endif
-	ne_md5_process_bytes(&pid, sizeof pid, hash);
+            ne_md5_process_bytes(&pid, sizeof pid, hash);
+        }
     }
 
-#ifdef HAVE_OPENSSL
-    }
-#endif
-
     ne_md5_finish_ascii(hash, ret);
     ne_md5_destroy_ctx(hash);
 
@@ -436,7 +452,8 @@ static void make_gss_error(ne_buffer *buf, int *flag,
 }
 
 /* Continue a GSS-API Negotiate exchange, using input TOKEN if
- * non-NULL.  Returns non-zero on error. */
+ * non-NULL.  Returns non-zero on error, in which case *errmsg is
+ * guaranteed to be non-NULL (i.e. an error message is set). */
 static int continue_negotiate(auth_session *sess, const char *token,
                               ne_buffer **errmsg)
 {
@@ -532,7 +549,7 @@ static int verify_negotiate_response(struct auth_request *req, auth_session *ses
     char *duphdr = ne_strdup(hdr);
     char *sep, *ptr = strchr(duphdr, ' ');
     int ret;
-    ne_buffer *errmsg;
+    ne_buffer *errmsg = NULL;
 
     if (strncmp(hdr, "Negotiate", ptr - duphdr) != 0) {
         ne_set_error(sess->sess, _("Negotiate response verification failed: "
@@ -560,8 +577,10 @@ static int verify_negotiate_response(struct auth_request *req, auth_session *ses
         ne_set_error(sess->sess, _("Negotiate response verification failure: %s"),
                      errmsg->data);
     }
-    ne_buffer_destroy(errmsg);
+
+    if (errmsg) ne_buffer_destroy(errmsg);
     ne_free(duphdr);
+
     return ret ? NE_ERROR : NE_OK;
 }
 #endif
@@ -609,6 +628,63 @@ static int sspi_challenge(auth_session *sess, int attempt,
 }
 #endif
 
+/* Parse the "domain" challenge parameter and set the domains array up
+ * in the session appropriately. */
+static int parse_domain(auth_session *sess, const char *domain)
+{
+    char *cp = ne_strdup(domain), *p = cp;
+    ne_uri base;
+    int invalid = 0;
+
+    memset(&base, 0, sizeof base);
+    ne_fill_server_uri(sess->sess, &base);
+
+    do {
+        char *token = ne_token(&p, ' ');
+        ne_uri rel, absolute;
+        
+        if (ne_uri_parse(token, &rel) == 0) {
+            /* Resolve relative to the Request-URI. */
+            ne_uri_resolve(&base, &rel, &absolute);
+
+            base.path = absolute.path;
+            
+            /* Ignore URIs not on this server. */
+            if (absolute.path && ne_uri_cmp(&absolute, &base) == 0) {
+                sess->domains = ne_realloc(sess->domains, 
+                                           ++sess->ndomains *
+                                           sizeof(*sess->domains));
+                sess->domains[sess->ndomains - 1] = absolute.path;
+                absolute.path = NULL;
+                NE_DEBUG(NE_DBG_HTTPAUTH, "auth: Using domain %s from %s\n",
+                         absolute.path, token);
+            }
+            else {
+                NE_DEBUG(NE_DBG_HTTPAUTH, "auth: Ignoring domain %s\n",
+                         token);
+            }
+
+            ne_uri_free(&absolute);
+        }
+        else {
+            invalid = 1;
+        }
+        
+        ne_uri_free(&rel);
+        
+    } while (p && !invalid);
+
+    if (invalid && sess->ndomains) {
+        free_domains(sess);
+    }
+
+    ne_free(cp);
+    base.path = NULL;
+    ne_uri_free(&base);
+
+    return invalid;
+}
+
 /* Examine a digest challenge: return 0 if it is a valid Digest challenge,
  * else non-zero. */
 static int digest_challenge(auth_session *sess, int attempt,
@@ -645,6 +721,14 @@ static int digest_challenge(auth_session *sess, int attempt,
         /* Non-stale challenge: clear session and request credentials. */
         clean_session(sess);
 
+        /* The domain paramater must be parsed after the session is
+         * cleaned; ignore domain for proxy auth. */
+        if (parms->domain && sess->spec == &ah_server_class
+            && parse_domain(sess, parms->domain)) {
+            challenge_error(errmsg, _("could not parse domain in Digest challenge"));
+            return -1;
+        }
+
         sess->realm = ne_strdup(parms->realm);
         sess->alg = parms->alg;
         sess->cnonce = get_cnonce();
@@ -718,6 +802,36 @@ static int digest_challenge(auth_session *sess, int attempt,
     return 0;
 }
 
+/* Returns non-zero if given Request-URI is inside the authentication
+ * domain defined for the session. */
+static int inside_domain(auth_session *sess, const char *req_uri)
+{
+    int inside = 0;
+    size_t n;
+    ne_uri uri;
+    
+    /* Parse the Request-URI; it will be an absoluteURI if using a
+     * proxy, and possibly '*'. */
+    if (strcmp(req_uri, "*") == 0 || ne_uri_parse(req_uri, &uri) != 0) {
+        /* Presume outside the authentication domain. */
+        return 0;
+    }
+
+    for (n = 0; n < sess->ndomains && !inside; n++) {
+        const char *d = sess->domains[n];
+        
+        inside = (d[strlen(d)-1] == '/'
+                  && strncmp(uri.path, d, strlen(d)) == 0)
+            || strcmp(d, uri.path) == 0;
+    }
+    
+    NE_DEBUG(NE_DBG_HTTPAUTH, "auth: '%s' is inside auth domain: %d.\n", 
+             uri.path, inside);
+    ne_uri_free(&uri);
+    
+    return inside;
+}            
+
 /* Return Digest authentication credentials header value for the given
  * session. */
 static char *request_digest(auth_session *sess, struct auth_request *req) 
@@ -728,6 +842,12 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
     const char *qop_value = "auth"; /* qop-value */
     ne_buffer *ret;
 
+    /* Do not submit credentials if an auth domain is defined and this
+     * request-uri fails outside it. */
+    if (sess->ndomains && !inside_domain(sess, req->uri)) {
+        return NULL;
+    }
+
     /* Increase the nonce-count */
     if (sess->qop != auth_qop_none) {
 	sess->nonce_count++;
@@ -1063,7 +1183,7 @@ static void challenge_error(ne_buffer **errbuf, const char *fmt, ...)
         ne_buffer_append(*errbuf, err, len);
     }
     else {
-        ne_buffer_concat(*errbuf, _(", "), err, NULL);
+        ne_buffer_concat(*errbuf, ", ", err, NULL);
     }
 }
 
@@ -1160,6 +1280,9 @@ static int auth_challenge(auth_session *sess, int attempt,
             
             chall->got_qop = chall->qop_auth;
 	}
+        else if (ne_strcasecmp(key, "domain") == 0) {
+            chall->domain = val;
+        }
     }
     
     sess->protocol = NULL;
@@ -1352,6 +1475,18 @@ static void auth_register(ne_session *sess, int isproxy, unsigned protomask,
     auth_session *ahs;
     struct auth_handler **hdl;
 
+    /* Handle the _ALL and _DEFAULT protocol masks: */
+    if (protomask == NE_AUTH_ALL) {
+        protomask |= NE_AUTH_BASIC | NE_AUTH_DIGEST | NE_AUTH_NEGOTIATE;
+    }
+    else if (protomask == NE_AUTH_DEFAULT) {
+        protomask |= NE_AUTH_BASIC | NE_AUTH_DIGEST;
+        
+        if (strcmp(ne_get_scheme(sess), "https") == 0 || isproxy) {
+            protomask |= NE_AUTH_NEGOTIATE;
+        }
+    }
+
     ahs = ne_get_session_private(sess, id);
     if (ahs == NULL) {
         ahs = ne_calloc(sizeof *ahs);
@@ -1403,29 +1538,16 @@ static void auth_register(ne_session *sess, int isproxy, unsigned protomask,
     (*hdl)->attempt = 0;
 }
 
-static void auth_register_default(ne_session *sess, int isproxy,
-                                  const struct auth_class *ahc, const char *id,
-                                  ne_auth_creds creds, void *userdata) 
-{
-    unsigned protomask = NE_AUTH_BASIC|NE_AUTH_DIGEST;
-    
-    if (strcmp(ne_get_scheme(sess), "https") == 0 || isproxy) {
-        protomask |= NE_AUTH_NEGOTIATE;
-    }
-
-    auth_register(sess, isproxy, protomask, ahc, id, creds, userdata);
-}
-
 void ne_set_server_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
 {
-    auth_register_default(sess, 0, &ah_server_class, HOOK_SERVER_ID,
-                          creds, userdata);
+    auth_register(sess, 0, NE_AUTH_DEFAULT, &ah_server_class, HOOK_SERVER_ID,
+                  creds, userdata);
 }
 
 void ne_set_proxy_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
 {
-    auth_register_default(sess, 1, &ah_proxy_class, HOOK_PROXY_ID,
-                          creds, userdata);
+    auth_register(sess, 1, NE_AUTH_DEFAULT, &ah_proxy_class, HOOK_PROXY_ID,
+                  creds, userdata);
 }
 
 void ne_add_server_auth(ne_session *sess, unsigned protocol, 
