@@ -1,6 +1,6 @@
 /*
    neon SSL/TLS support using GNU TLS
-   Copyright (C) 2002-2007, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2008, Joe Orton <joe@manyfish.co.uk>
    Copyright (C) 2004, Aleix Conchillo Flaque <aleix@member.fsf.org>
 
    This library is free software; you can redistribute it and/or
@@ -17,13 +17,10 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
-
 */
 
 #include "config.h"
 
-#include <sys/mman.h>
-#include <sys/stat.h>
 #include <sys/types.h>
 
 #ifdef HAVE_STRING_H
@@ -72,6 +69,7 @@ struct ne_ssl_certificate_s {
 struct ne_ssl_client_cert_s {
     gnutls_pkcs12 p12;
     int decrypted; /* non-zero if successfully decrypted. */
+    int keyless;
     ne_ssl_certificate cert;
     gnutls_x509_privkey pkey;
     char *friendly_name;
@@ -331,7 +329,6 @@ void ne_ssl_cert_validity_time(const ne_ssl_certificate *cert,
 static int match_hostname(char *cn, const char *hostname)
 {
     const char *dot;
-    NE_DEBUG(NE_DBG_SSL, "Match %s on %s...\n", cn, hostname);
     dot = strchr(hostname, '.');
     if (dot == NULL) {
 	char *pnt = strchr(cn, '.');
@@ -445,7 +442,10 @@ static int check_identity(const ne_uri *server, gnutls_x509_crt cert,
         }
     }
 
-    NE_DEBUG(NE_DBG_SSL, "Identity match: %s\n", match ? "good" : "bad");
+    if (*hostname)
+        NE_DEBUG(NE_DBG_SSL, "ssl: Identity match for '%s': %s\n", hostname, 
+                 match ? "good" : "bad");
+
     return match ? 0 : 1;
 }
 
@@ -505,13 +505,18 @@ static ne_ssl_client_cert *dup_client_cert(const ne_ssl_client_cert *cc)
     ne_ssl_client_cert *newcc = ne_calloc(sizeof *newcc);
 
     newcc->decrypted = 1;
-
-    ret = gnutls_x509_privkey_init(&newcc->pkey);
-    if (ret != 0) goto dup_error;
-
-    ret = gnutls_x509_privkey_cpy(newcc->pkey, cc->pkey);
-    if (ret != 0) goto dup_error;
     
+    if (cc->keyless) {
+        newcc->keyless = 1;
+    }
+    else {
+        ret = gnutls_x509_privkey_init(&newcc->pkey);
+        if (ret != 0) goto dup_error;
+        
+        ret = gnutls_x509_privkey_cpy(newcc->pkey, cc->pkey);
+        if (ret != 0) goto dup_error;
+    }    
+
     newcc->cert.subject = x509_crt_copy(cc->cert.subject);
     if (!newcc->cert.subject) goto dup_error;
 
@@ -536,7 +541,7 @@ static int provide_client_cert(gnutls_session session,
     ne_session *sess = gnutls_session_get_ptr(session);
     
     if (!sess) {
-        return -1;
+        return GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER;
     }
 
     if (!sess->client_cert && sess->ssl_provide_fn) {
@@ -561,10 +566,11 @@ static int provide_client_cert(gnutls_session session,
             /* tell GNU TLS not to deallocate the certs. */
             st->deinit_all = 0;
         } else {
-            return -1;
+            return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
         }
     } else {
         NE_DEBUG(NE_DBG_SSL, "No client certificate supplied.\n");
+        return GNUTLS_E_NO_CERTIFICATE_FOUND;
     }
 
     return 0;
@@ -717,6 +723,10 @@ int ne__negotiate_ssl(ne_session *sess)
 
     NE_DEBUG(NE_DBG_SSL, "Negotiating SSL connection.\n");
 
+    /* Pass through the hostname if SNI is enabled. */
+    ctx->hostname = 
+        sess->flags[NE_SESSFLAG_TLS_SNI] ? sess->server.hostname : NULL;
+
     if (ne_sock_connect_ssl(sess->socket, ctx, sess)) {
 	ne_set_error(sess, _("SSL negotiation failed: %s"),
 		     ne_sock_error(sess->socket));
@@ -915,7 +925,8 @@ ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
     }
 
     if (gnutls_pkcs12_verify_mac(p12, "") == 0) {
-        if (pkcs12_parse(p12, &pkey, &cert, &friendly_name, "") != 0) {
+        if (pkcs12_parse(p12, &pkey, &cert, &friendly_name, "") != 0
+            || !cert || !pkey) {
             gnutls_pkcs12_deinit(p12);
             return NULL;
         }
@@ -937,6 +948,30 @@ ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
     }
 }
 
+ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
+                                                 size_t der_len)
+{
+    ne_ssl_client_cert *cc;
+    gnutls_x509_crt x5;
+    gnutls_datum datum;
+
+    datum.data = (unsigned char *)der;
+    datum.size = der_len;    
+
+    if (gnutls_x509_crt_init(&x5) 
+        || gnutls_x509_crt_import(x5, &datum, GNUTLS_X509_FMT_DER)) {
+        NE_DEBUG(NE_DBG_SSL, "ssl: crt_import failed.\n");
+        return NULL;
+    }
+    
+    cc = ne_calloc(sizeof *cc);
+    cc->keyless = 1;
+    cc->decrypted = 1;
+    populate_cert(&cc->cert, x5);
+
+    return cc;    
+}
+
 int ne_ssl_clicert_encrypted(const ne_ssl_client_cert *cc)
 {
     return !cc->decrypted;
@@ -955,6 +990,12 @@ int ne_ssl_clicert_decrypt(ne_ssl_client_cert *cc, const char *password)
     ret = pkcs12_parse(cc->p12, &pkey, &cert, NULL, password);
     if (ret < 0)
         return ret;
+    
+    if (!cert || (!pkey && !cc->keyless)) {
+        if (cert) gnutls_x509_crt_deinit(cert);
+        if (pkey) gnutls_x509_privkey_deinit(pkey);
+        return -1;
+    }
 
     gnutls_pkcs12_deinit(cc->p12);
     populate_cert(&cc->cert, cert);
