@@ -1,6 +1,6 @@
 /* 
    Socket handling routines
-   Copyright (C) 1998-2007, Joe Orton <joe@manyfish.co.uk>, 
+   Copyright (C) 1998-2008, Joe Orton <joe@manyfish.co.uk>, 
    Copyright (C) 1999-2000 Tommi Komulainen <Tommi.Komulainen@iki.fi>
    Copyright (C) 2004 Aleix Conchillo Flaque <aleix@member.fsf.org>
 
@@ -192,9 +192,13 @@ struct iofns {
     int (*readable)(ne_socket *s, int n);
 };
 
+static const ne_inet_addr dummy_laddr;
+
 struct ne_socket_s {
     int fd;
-    char error[200];
+    unsigned int lport;
+    const ne_inet_addr *laddr;
+
     void *progress_ud;
     int rdtimeout, cotimeout; /* timeouts */
     const struct iofns *ops;
@@ -206,10 +210,12 @@ struct ne_socket_s {
      * advances through ->buffer.  ->bufavail gives the number of
      * bytes which remain to be consumed in ->buffer (from ->bufpos),
      * and is hence always <= RDBUFSIZ. */
-#define RDBUFSIZ 4096
-    char buffer[RDBUFSIZ];
     char *bufpos;
     size_t bufavail;
+#define RDBUFSIZ 4096
+    char buffer[RDBUFSIZ];
+    /* Error string. */
+    char error[192];
 };
 
 /* ne_sock_addr represents an Internet address. */
@@ -1105,18 +1111,82 @@ ne_socket *ne_sock_create(void)
     return sock;
 }
 
+
+#ifdef USE_GETADDRINFO
+#define ia_family(a) ((a)->ai_family)
+#define ia_proto(a)  ((a)->ai_protocol)
+#else
+#define ia_family(a) AF_INET
+#define ia_proto(a)  0
+#endif
+
+void ne_sock_prebind(ne_socket *sock, const ne_inet_addr *addr,
+                     unsigned int port)
+{
+    sock->lport = port;
+    sock->laddr = addr ? addr : &dummy_laddr;    
+}
+
+/* Bind socket 'fd' to address/port 'addr' and 'port', for subsequent
+ * connect() to address of family 'peer_family'. */
+static int do_bind(int fd, int peer_family, 
+                   const ne_inet_addr *addr, unsigned int port)
+{
+#if defined(HAVE_SETSOCKOPT) && defined(SO_REUSEADDR) && defined(SOL_SOCKET)
+    {
+        int flag = 1;
+
+        (void) setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof flag);
+        /* An error here is not fatal, so ignore it. */
+    }
+#endif        
+    
+
+#ifdef USE_GETADDRINFO
+    /* Use a sockaddr_in6 if an AF_INET6 local address is specifed, or
+     * if no address is specified and the peer address is AF_INET6: */
+    if ((addr != &dummy_laddr && addr->ai_family == AF_INET6)
+        || (addr == &dummy_laddr && peer_family == AF_INET6)) {
+        struct sockaddr_in6 in6;
+        
+        if (addr == &dummy_laddr)
+            memset(&in6, 0, sizeof in6);
+        else
+            memcpy(&in6, addr->ai_addr, sizeof in6);
+        in6.sin6_port = htons(port);
+        /* fill in the _family field for AIX 4.3, which forgets to do so. */
+        in6.sin6_family = AF_INET6;
+
+        return bind(fd, (struct sockaddr *)&in6, sizeof in6);
+    } else
+#endif
+    {
+	struct sockaddr_in in;
+
+        if (addr == &dummy_laddr)
+            memset(&in, 0, sizeof in);
+        else {
+#ifdef USE_GETADDRINFO
+            memcpy(&in, addr->ai_addr, sizeof in);
+#else
+            in.sin_addr = *addr;
+#endif
+        }
+        in.sin_port = htons(port);
+        in.sin_family = AF_INET;
+
+        return bind(fd, (struct sockaddr *)&in, sizeof in);
+    }
+}
+
 int ne_sock_connect(ne_socket *sock,
                     const ne_inet_addr *addr, unsigned int port)
 {
     int fd, ret;
 
-#ifdef USE_GETADDRINFO
     /* use SOCK_STREAM rather than ai_socktype: some getaddrinfo
      * implementations do not set ai_socktype, e.g. RHL6.2. */
-    fd = socket(addr->ai_family, SOCK_STREAM, addr->ai_protocol);
-#else
-    fd = socket(AF_INET, SOCK_STREAM, 0);
-#endif
+    fd = socket(ia_family(addr), SOCK_STREAM, ia_proto(addr));
     if (fd < 0) {
         set_strerror(sock, ne_errno);
 	return -1;
@@ -1129,6 +1199,26 @@ int ne_sock_connect(ne_socket *sock,
         return NE_SOCK_ERROR;
     }
 #endif
+   
+#if defined(HAVE_FCNTL) && defined(F_GETFD) && defined(F_SETFD) \
+    && defined(FD_CLOEXEC)
+    /* Set the FD_CLOEXEC bit for the new fd. */
+    if ((ret = fcntl(fd, F_GETFD)) >= 0) {
+        fcntl(fd, F_SETFD, ret | FD_CLOEXEC);
+        /* ignore failure; not a critical error. */
+    }
+#endif
+
+    if (sock->laddr && (sock->laddr == &dummy_laddr || 
+                        ia_family(sock->laddr) == ia_family(addr))) {
+        ret = do_bind(fd, ia_family(addr), sock->laddr, sock->lport);
+        if (ret < 0) {
+            int errnum = errno;
+            ne_close(fd);
+            set_strerror(sock, errnum);
+            return NE_SOCK_ERROR;
+        }
+    }
 
 #if defined(TCP_NODELAY) && defined(HAVE_SETSOCKOPT) && defined(IPPROTO_TCP)
     { /* Disable the Nagle algorithm; better to add write buffering
@@ -1141,10 +1231,49 @@ int ne_sock_connect(ne_socket *sock,
     ret = connect_socket(sock, fd, addr, htons(port));
     if (ret == 0)
         sock->fd = fd;
+    else
+        ne_close(fd);
 
     return ret;
 }
 
+ne_inet_addr *ne_sock_peer(ne_socket *sock, unsigned int *port)
+{
+    union saun {
+        struct sockaddr_in sin;
+#ifdef USE_GETADDRINFO
+        struct sockaddr_in6 sin6;
+#endif
+    } saun;
+    socklen_t len = sizeof saun;
+    ne_inet_addr *ia;
+    struct sockaddr *sad = (struct sockaddr *)&saun;
+
+    if (getpeername(sock->fd, sad, &len) != 0) {
+        set_strerror(sock, errno);
+        return NULL;
+    }
+
+    ia = ne_calloc(sizeof *ia);
+#ifdef USE_GETADDRINFO
+    ia->ai_addr = ne_malloc(sizeof *ia);
+    ia->ai_addrlen = len;
+    memcpy(ia->ai_addr, sad, len);
+    ia->ai_family = sad->sa_family;
+#else
+    memcpy(ia, &saun.sin.sin_addr.s_addr, sizeof *ia);
+#endif    
+
+#ifdef USE_GETADDRINFO
+    *port = ntohs(sad->sa_family == AF_INET ? 
+                  saun.sin.sin_port : saun.sin6.sin6_port);
+#else
+    *port = ntohs(saun.sin.sin_port);
+#endif
+
+    return ia;
+}
+
 ne_inet_addr *ne_iaddr_make(ne_iaddr_type type, const unsigned char *raw)
 {
     ne_inet_addr *ia;
@@ -1333,7 +1462,7 @@ int ne_sock_accept_ssl(ne_socket *sock, ne_ssl_context *ctx)
         gnutls_certificate_server_set_request(ssl, GNUTLS_CERT_REQUEST);
 
     sock->ssl = ssl;
-    gnutls_transport_set_ptr(sock->ssl, (gnutls_transport_ptr) sock->fd);
+    gnutls_transport_set_ptr(sock->ssl, (gnutls_transport_ptr)(long)sock->fd);
     ret = gnutls_handshake(ssl);
     if (ret < 0) {
         return error_gnutls(sock, ret);
@@ -1376,6 +1505,17 @@ int ne_sock_connect_ssl(ne_socket *sock, ne_ssl_context *ctx, void *userdata)
     SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
     SSL_set_fd(ssl, sock->fd);
     sock->ops = &iofns_ssl;
+
+#ifdef SSL_set_tlsext_host_name
+    if (ctx->hostname) {
+        /* Try to enable SNI, but ignore failure (should only fail for
+         * >255 char hostnames, which are probably not legal
+         * anyway).  */
+        if (SSL_set_tlsext_host_name(ssl, ctx->hostname) != 1) {
+            ERR_clear_error();
+        }
+    }
+#endif
     
     if (ctx->sess)
 	SSL_set_session(ssl, ctx->sess);
@@ -1394,7 +1534,17 @@ int ne_sock_connect_ssl(ne_socket *sock, ne_ssl_context *ctx, void *userdata)
     gnutls_session_set_ptr(sock->ssl, userdata);
     gnutls_credentials_set(sock->ssl, GNUTLS_CRD_CERTIFICATE, ctx->cred);
 
-    gnutls_transport_set_ptr(sock->ssl, (gnutls_transport_ptr) sock->fd);
+#ifdef HAVE_GNUTLS_SIGN_CALLBACK_SET
+    if (ctx->sign_func)
+        gnutls_sign_callback_set(sock->ssl, ctx->sign_func, ctx->sign_data);    
+#endif
+
+    if (ctx->hostname) {
+        gnutls_server_name_set(sock->ssl, GNUTLS_NAME_DNS, ctx->hostname,
+                               strlen(ctx->hostname));
+    }                               
+
+    gnutls_transport_set_ptr(sock->ssl, (gnutls_transport_ptr)(long)sock->fd);
 
     if (ctx->cache.client.data) {
 #if defined(HAVE_GNUTLS_SESSION_GET_DATA2)
@@ -1468,8 +1618,8 @@ int ne_sock_sessid(ne_socket *sock, unsigned char *buf, size_t *buflen)
         return -1;
     }
 
-    memcpy(buf, sess->session_id, *buflen);
     *buflen = sess->session_id_length;
+    memcpy(buf, sess->session_id, *buflen);
     return 0;
 #endif
 #else
@@ -1477,6 +1627,25 @@ int ne_sock_sessid(ne_socket *sock, unsigned char *buf, size_t *buflen)
 #endif
 }
 
+char *ne_sock_cipher(ne_socket *sock)
+{
+#ifdef NE_HAVE_SSL
+    if (sock->ssl) {
+#ifdef HAVE_OPENSSL
+        const char *name = SSL_get_cipher(sock->ssl);
+        return ne_strdup(name);
+#elif defined(HAVE_GNUTLS)
+        const char *name = gnutls_cipher_get_name(gnutls_cipher_get(sock->ssl));
+        return ne_strdup(name);
+#endif
+    }
+    else 
+#endif /* NE_HAVE_SSL */
+    {
+        return NULL;
+    }    
+}
+
 const char *ne_sock_error(const ne_socket *sock)
 {
     return sock->error;
